--  Copyright (C) 1994-2001 Grady Booch and Simon Wright.
--  All Rights Reserved.
--
--      This program is free software; you can redistribute it
--      and/or modify it under the terms of the Ada Community
--      License which comes with this Library.
--
--      This program is distributed in the hope that it will be
--      useful, but WITHOUT ANY WARRANTY; without even the implied
--      warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
--      PURPOSE. See the Ada Community License for more details.
--      You should have received a copy of the Ada Community
--      License with this library, in the file named "Ada Community
--      License" or "ACL". If not, contact the author of this library
--      for a copy.
--

--  $RCSfile: bc-containers-queues-synchronized.adb,v $
--  $Revision: 1.9 $
--  $Date: 2001/09/22 17:37:08 $
--  $Author: simon $

with System.Address_To_Access_Conversions;

package body BC.Containers.Queues.Synchronized is

   procedure Clear (Q : in out Queue) is
      Region : BC.Support.Synchronization.Lock (Q.Critical_Region);
      pragma Warnings (Off, Region);
      Lock : BC.Support.Synchronization.Write_Lock (Q.The_Monitor);
      pragma Warnings (Off, Lock);
   begin
      if Length (Queue_Base (Q)) > 0 then
         BC.Support.Synchronization.Seize (Q.Producer_Consumer.all);
      end if;
      Clear (Queue_Base (Q));
   end Clear;

   procedure Append (Q : in out Queue; Elem : Item) is
      Lock : BC.Support.Synchronization.Write_Lock (Q.The_Monitor);
      pragma Warnings (Off, Lock);
   begin
      if Length (Queue_Base (Q)) = 0 then
         BC.Support.Synchronization.Release (Q.Producer_Consumer.all);
      end if;
      Append (Queue_Base (Q), Elem);
   end Append;

   procedure Pop (Q : in out Queue) is
      Region : BC.Support.Synchronization.Lock (Q.Critical_Region);
      pragma Warnings (Off, Region);
   begin
      BC.Support.Synchronization.Seize (Q.Producer_Consumer.all);
      declare
         Lock : BC.Support.Synchronization.Write_Lock (Q.The_Monitor);
         pragma Warnings (Off, Lock);
      begin
         if Length (Queue_Base (Q)) > 1 then
            BC.Support.Synchronization.Release (Q.Producer_Consumer.all);
         end if;
         Pop (Queue_Base (Q));
      end;
   end Pop;

   procedure Pop_Value (Q : in out Queue; Elem : out Item) is
      Region : BC.Support.Synchronization.Lock (Q.Critical_Region);
      pragma Warnings (Off, Region);
   begin
      BC.Support.Synchronization.Seize (Q.Producer_Consumer.all);
      declare
         Lock : BC.Support.Synchronization.Write_Lock (Q.The_Monitor);
         pragma Warnings (Off, Lock);
      begin
         if Length (Queue_Base (Q)) > 1 then
            BC.Support.Synchronization.Release (Q.Producer_Consumer.all);
         end if;
         Elem := Front (Queue_Base (Q));
         Pop (Queue_Base (Q));
      end;
   end Pop_Value;

   procedure Remove (Q : in out Queue; From : Positive) is
      Region : BC.Support.Synchronization.Lock (Q.Critical_Region);
      pragma Warnings (Off, Region);
   begin
      BC.Support.Synchronization.Seize (Q.Producer_Consumer.all);
      declare
         Lock : BC.Support.Synchronization.Write_Lock (Q.The_Monitor);
         pragma Warnings (Off, Lock);
      begin
         if Length (Queue_Base (Q)) > 1 then
            BC.Support.Synchronization.Release (Q.Producer_Consumer.all);
         end if;
         Remove (Queue_Base (Q), From);
      end;
   end Remove;

   function Available (Q : in Queue) return Natural is
      Lock : BC.Support.Synchronization.Read_Lock (Q.The_Monitor);
      pragma Warnings (Off, Lock);
   begin
      return Available (Queue_Base (Q));
   end Available;

   function Length (Q : in Queue) return Natural is
      Lock : BC.Support.Synchronization.Read_Lock (Q.The_Monitor);
      pragma Warnings (Off, Lock);
   begin
      return Length (Queue_Base (Q));
   end Length;

   function Is_Empty (Q : in Queue) return Boolean is
      Lock : BC.Support.Synchronization.Read_Lock (Q.The_Monitor);
      pragma Warnings (Off, Lock);
   begin
      return Is_Empty (Queue_Base (Q));
   end Is_Empty;

   function Front (Q : in Queue) return Item is
      Lock : BC.Support.Synchronization.Read_Lock (Q.The_Monitor);
      pragma Warnings (Off, Lock);
   begin
      return Front (Queue_Base (Q));
   end Front;

   function Location
     (Q : in Queue; Elem : in Item) return Natural is
      Lock : BC.Support.Synchronization.Read_Lock (Q.The_Monitor);
      pragma Warnings (Off, Lock);
   begin
      return Location (Queue_Base (Q), Elem);
   end Location;

   function "=" (Left, Right : in Queue) return Boolean is
      --  To avoid deadlock, make sure we always apply locks in the same order
      use type System.Address;
   begin
      if Left'Address = Right'Address then
         return True;
      elsif Left'Address > Right'Address then
         return Right = Left;
      else
         declare
            Left_Lock :
              BC.Support.Synchronization.Read_Lock (Left.The_Monitor);
            pragma Warnings (Off, Left_Lock);
            Right_Lock :
              BC.Support.Synchronization.Read_Lock (Right.The_Monitor);
            pragma Warnings (Off, Right_Lock);
         begin
            return Queue_Base (Left) = Queue_Base (Right);
         end;
      end if;
   end "=";

   package Address_Conversions
   is new System.Address_To_Access_Conversions (Queue);

   function New_Iterator
     (For_The_Queue : Queue) return Iterator'Class is
      Result : Synchronized_Queue_Iterator;
   begin
      Result.For_The_Container :=
        Address_Conversions.To_Pointer (For_The_Queue'Address).all'Access;
      Reset (Result);
      Lock (Result.For_The_Container.all);
      return Result;
   end New_Iterator;

   procedure Lock (Q : in out Queue) is
   begin
      BC.Support.Synchronization.Seize_For_Reading (Q.The_Monitor.all);
   end Lock;

   procedure Unlock (Q : in out Queue) is
   begin
      BC.Support.Synchronization.Release_From_Reading (Q.The_Monitor.all);
   end Unlock;

   procedure Initialize (Q : in out Queue) is
   begin
      Initialize (Queue_Base (Q));
      Q.The_Monitor := new Monitor;
      Q.Critical_Region := new BC.Support.Synchronization.Semaphore;
      Q.Producer_Consumer := new BC.Support.Synchronization.Semaphore;
      BC.Support.Synchronization.Seize (Q.Producer_Consumer.all);
   end Initialize;

   procedure Adjust (Q : in out Queue) is
      --  Consider Q := P;
      --  On entry to Adjust, the non-controlled parts of Q contains a
      --  bitwise copy of P, so that Q.The_Monitor points to the same
      --  Monitor as P.The_Monitor.
   begin
      --  lock P's monitor
      Lock (Q);
      --  make the deep copy
      Adjust (Queue_Base (Q));
      --  unlock P's monitor
      Unlock (Q);
      --  create new monitor, semaphores for Q
      Q.The_Monitor := new Monitor;
      Q.Critical_Region := new BC.Support.Synchronization.Semaphore;
      Q.Producer_Consumer := new BC.Support.Synchronization.Semaphore;
   end Adjust;

   procedure Finalize (Q : in out Queue) is
   begin
      BC.Support.Synchronization.Delete (Q.The_Monitor);
      BC.Support.Synchronization.Delete (Q.Critical_Region);
      BC.Support.Synchronization.Delete (Q.Producer_Consumer);
      Finalize (Queue_Base (Q));
   end Finalize;

   Empty_Container : Queue;
   pragma Warnings (Off, Empty_Container);

   function Null_Container return Queue is
   begin
      return Empty_Container;
   end Null_Container;

   procedure Adjust (The_Iterator : in out Synchronized_Queue_Iterator) is
   begin
      Lock (The_Iterator.For_The_Container.all);
   end Adjust;

   procedure Finalize (The_Iterator : in out Synchronized_Queue_Iterator) is
   begin
      Unlock (The_Iterator.For_The_Container.all);
   end Finalize;

end BC.Containers.Queues.Synchronized;
