--  Copyright (C) 1994-2001 Grady Booch and Simon Wright.
--  All Rights Reserved.
--
--      This program is free software; you can redistribute it
--      and/or modify it under the terms of the Ada Community
--      License which comes with this Library.
--
--      This program is distributed in the hope that it will be
--      useful, but WITHOUT ANY WARRANTY; without even the implied
--      warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
--      PURPOSE. See the Ada Community License for more details.
--      You should have received a copy of the Ada Community
--      License with this library, in the file named "Ada Community
--      License" or "ACL". If not, contact the author of this library
--      for a copy.
--

--  $RCSfile: bc-containers-queues-synchronized.ads,v $
--  $Revision: 1.5 $
--  $Date: 2001/09/22 17:37:08 $
--  $Author: simon $

with BC.Support.Synchronization;

generic
   type Queue_Base is new Abstract_Queue with private;
   type Monitor is new BC.Support.Synchronization.Monitor_Base with private;
package BC.Containers.Queues.Synchronized is

   pragma Elaborate_Body;

   --  A queue denotes a sequence of items, in which items may be
   --  added from one end and removed from the opposite end of the
   --  sequence.

   type Queue is new Queue_Base with private;

   function Null_Container return Queue;

   procedure Clear (Q : in out Queue);
   --  Empty the queue of all items.

   procedure Append (Q : in out Queue; Elem : Item);
   --  Add the item to the back of the queue; the item itself is
   --  copied.

   procedure Pop (Q : in out Queue);
   --  Remove the item from the front of the queue.

   procedure Pop_Value (Q : in out Queue; Elem : out Item);
   --  Remove and return the item from the front of the queue.

   procedure Remove (Q : in out Queue; From : Positive);
   --  Remove the item at the given index (may be a balking
   --  operation).

   function Available (Q : in Queue) return Natural;
   --  Indicates number of empty "Item slots" left in Queue

   function Length (Q : in Queue) return Natural;
   --  Return the number of items in the queue.

   function Is_Empty (Q : in Queue) return Boolean;
   --  Return True if and only if there are no items in the queue.

   function Front (Q : in Queue) return Item;
   --  Return a copy of the item at the front of the queue.

   function Location (Q : in Queue; Elem : in Item) return Natural;
   --  Return the first index at which the item is found; return 0 if
   --  the item does not exist in the queue.

   function "=" (Left, Right : in Queue) return Boolean;
   --  Return True if and only if both queues have the same length and
   --  the same items in the same order; return False otherwise.

   function New_Iterator
     (For_The_Queue : Queue) return Iterator'Class;
   --  Return a reset Iterator bound to the specific Queue.

private

   procedure Lock (Q : in out Queue);
   procedure Unlock (Q : in out Queue);

   type Queue is new Queue_Base with record
      The_Monitor : BC.Support.Synchronization.Monitor_P;
      Critical_Region : BC.Support.Synchronization.Semaphore_P;
      Producer_Consumer : BC.Support.Synchronization.Semaphore_P;
      --  Bounded_Buffer : BC.Support.Synchronization.Semaphore_P;
      --  In the C++, Bounded_Buffer only gets used for the Bounded form
   end record;

   procedure Initialize (Q : in out Queue);
   procedure Adjust (Q : in out Queue);
   procedure Finalize (Q : in out Queue);

   type Synchronized_Queue_Iterator is new Queue_Iterator with null record;

   --  There is no Initialize since the only way of obtaining a
   --  Synchronized_Queue_Iterator is via New_Interator, which
   --  performs the necessary magic.
   procedure Adjust (The_Iterator : in out Synchronized_Queue_Iterator);
   procedure Finalize (The_Iterator : in out Synchronized_Queue_Iterator);

end BC.Containers.Queues.Synchronized;
